/* -----------------------------------------------------------------------------
*
*  npkey.c
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2012 by Shane Saxon - contact@saxondigital.net
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named COPYING.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#include "npkey.h"

#include <freeglut.h>

#include "../npdata.h"
#include "../npctrl.h"


void npKeyEvent (int key, int keyEventType);

//------------------------------------------------------------------------------
void npInitKey (void* dataRef)
{
	int i = 0;
	int j = 0;

	pData data = (pData) dataRef;
	pNPkey key = &data->io.key;

	key->modShift = 0;
	key->modCtrl = 0;
	key->modAlt = 0;
	key->modCommand = 0;

	for (i = 0; i < kNPkeyEventTypeSize; i++)
		for (j = 0; j < kNPkeyMapSize; j++)
			key->map[i][j] = 0;


	//the key map defines which command to execute for each key event
	//three types of fundamental keys, toggle, momentary, repeating
	//four event types generated, keyDown, KeyRepeat, KeyUp, KeyModifier

	//Fire keys, respond to keyDown only
	key->map[kKeyDown][kKeyCodeM]				= kNPgraphMenu;

	key->map[kKeyDown][kKeyCodeY]				= kNPgraphSegments;
	key->map[kKeyDown][kKeyCodeT]				= kNPgraphTexture;

	key->map[kKeyDown][kKeyCodeESC]				= kFullScreenExit;
//	key->map[kKeyDown][kKeyCodeB]				= kNPgraphClearFrameBuffer;
	key->map[kKeyDown][kKeyCodeB]				= kNPgraphBackground;
	key->map[kKeyDown][kKeyCodeX]				= kNPgraphSelectAxes;

	key->map[kKeyDown][kKeyCodeC]				= kNPgraphCamera;
	key->map[kKeyDown][kKeyCodeG]				= kNPgraphGrid;

	key->map[kKeyDown][kKeyCodeSpace]			= kNPgraphSelect;
	key->map[kKeyDown][kKeyCode4]				= kNPgraphSelectAll;
	key->map[kKeyDown][kKeyCodeTilda]			= kNPgraphSelectAllOff;
	key->map[kKeyDown][kKeyCode1]				= kNPfileMapOne;
	key->map[kKeyDown][kKeyCode2]				= kNPfileMapTwo;
	key->map[kKeyDown][kKeyCode3]				= kNPfileMapThree;

	key->map[kKeyDown][kKeyCode5]				= kNPgraphPresetOne;
	key->map[kKeyDown][kKeyCode6]				= kNPgraphPresetTwo;
	key->map[kKeyDown][kKeyCode7]				= kNPgraphPresetThree;
//	key->map[kKeyDown][kKeyCode8]				= kNPgraphPresetFour;

//	key->map[kKeyDown][kKeyCodeC]				= kNPgraphCenter;
	key->map[kKeyDown][kKeyCodeBackSlash]		= kNPgraphScroll;
	key->map[kKeyDown][kKeyCodeF]				= kNPgraphFreeze;
	key->map[kKeyDown][kKeyCodeH]				= kNPgraphHide;

	key->map[kKeyDown][kKeyCodeO]				= kNPgraphPrimitive;
	key->map[kKeyDown][kKeyCodeJ]				= kNPgraphTopo;

	key->map[kKeyDown][kKeyCodeI]				= kNPgraphTagType;
	
	key->map[kKeyDown][kKeyCodeK]				= kNPfileSave;
	key->map[kKeyDown][kKeyCodeL]				= kNPfileOpen;
	key->map[kKeyDown][kKeyCodeP]				= kNPfileImport;	// kNPgraphPoints;

	key->map[kKeyDown][kKeyCodeU]				= kNPopenURL;


//	key->map[kKeyDown][kKeyCodeL]				= kNPgraphMesh;
	

	key->map[kKeyDown][kKeyCodeLeftBracket]		= kEventLowSetpoint;
	key->map[kKeyDown][kKeyCodeRightBracket]	= kEventHighSetpoint;

	//Repeat keys, don't usually respond to keyUp
	key->map[kKeyDown][kKeyCodeN]				= kNPgraphNew;
	key->map[kKeyRepeat][kKeyCodeN]				= kNPgraphNew;
	
	key->map[kKeyDown][kKeyCodeDelete]			= kNPgraphDelete;
	key->map[kKeyRepeat][kKeyCodeDelete]		= kNPgraphDelete;

	key->map[kKeyDown][kKeyCodeBackSpace]		= kNPgraphDelete;
	key->map[kKeyRepeat][kKeyCodeBackSpace]		= kNPgraphDelete;

	key->map[kKeyDown][kKeyCodeTab]				= kNPgraphNext;
	key->map[kKeyRepeat][kKeyCodeTab]			= kNPgraphNext;

	key->map[kKeyDown][kKeyCodeReturn]			= kNPgraphNextBranch;
	key->map[kKeyRepeat][kKeyCodeReturn]		= kNPgraphNextBranch;
	key->map[kKeyDown][kKeyCodeNumPadEnter]		= kNPgraphNextBranch;
	key->map[kKeyRepeat][kKeyCodeNumPadEnter]	= kNPgraphNextBranch;

	key->map[kKeyDown][kKeyCodeComma]			= kNPgraphChannelDown;
	key->map[kKeyRepeat][kKeyCodeComma]			= kNPgraphChannelDown;
	
	key->map[kKeyDown][kKeyCodePeriod]			= kNPgraphChannelUp;
	key->map[kKeyRepeat][kKeyCodePeriod]		= kNPgraphChannelUp;

	key->map[kKeyDown][kKeyCodeMinus]			= kNPgraphColorDown;
	key->map[kKeyRepeat][kKeyCodeMinus]			= kNPgraphColorDown;
	
	key->map[kKeyDown][kKeyCodeEqual]			= kNPgraphColorUp;
	key->map[kKeyRepeat][kKeyCodeEqual]			= kNPgraphColorUp;

	key->map[kKeyDown][kKeyCodeZero]			= kNPgraphAlphaUp;
	key->map[kKeyRepeat][kKeyCodeZero]			= kNPgraphAlphaUp;

	key->map[kKeyDown][kKeyCode9]				= kNPgraphAlphaDown;
	key->map[kKeyRepeat][kKeyCode9]				= kNPgraphAlphaDown;

	key->map[kKeyDown][kKeyCode8]				= kNPgraphAlphaMode;
	key->map[kKeyRepeat][kKeyCode8]				= kNPgraphAlphaMode;

	key->map[kKeyDown][kKeyCodeR]				= kNPgraphRatio; //kNPgraphNormalMode;
	key->map[kKeyRepeat][kKeyCodeR]				= kNPgraphRatio; //kNPgraphNormalMode;

	

//	key->map[kKeyDown][kKeyCode]				= kNPgraphShaderMode;
//	key->map[kKeyRepeat][kKeyCode]				= kNPgraphShaderMode;


//	key->map[kKeyDown][kKeyCode]				= kNPgraphBiasUp;
//	key->map[kKeyRepeat][kKeyCode]				= kNPgraphBiasUp;
//	key->map[kKeyDown][kKeyCode]				= kNPgraphBiasDown;
//	key->map[kKeyRepeat][kKeyCode]				= kNPgraphBiasDown;

//	key->map[kKeyDown][kKeyCode]				= kNPgraphGainUp;
//	key->map[kKeyRepeat][kKeyCode]				= kNPgraphGainUp;
//	key->map[kKeyDown][kKeyCode]				= kNPgraphGainDown;
//	key->map[kKeyRepeat][kKeyCode]				= kNPgraphGainDown;

	//Momentary Keys are active while pressed and disable upon keyup
	//Arrow keys for X,Y,Z translate, rotate and scale of selected objects

	// translate
	key->map[kKeyDown][kKeyCodeA]				= kNPgraphXdecrease;
	key->map[kKeyUp][kKeyCodeA]					= kNPgraphXdecreaseOff;

	key->map[kKeyDown][kKeyCodeD]				= kNPgraphXincrease;
	key->map[kKeyUp][kKeyCodeD]					= kNPgraphXincreaseOff;

	key->map[kKeyDown][kKeyCodeW]				= kNPgraphYincrease;
	key->map[kKeyUp][kKeyCodeW]					= kNPgraphYincreaseOff;

	key->map[kKeyDown][kKeyCodeS]				= kNPgraphYdecrease;
	key->map[kKeyUp][kKeyCodeS]					= kNPgraphYdecreaseOff;

	key->map[kKeyDown][kKeyCodeE]				= kNPgraphZincrease;
	key->map[kKeyUp][kKeyCodeE]					= kNPgraphZincreaseOff;

	key->map[kKeyDown][kKeyCodeQ]				= kNPgraphZdecrease;
	key->map[kKeyUp][kKeyCodeQ]					= kNPgraphZdecreaseOff;

	// rotate
	key->map[kKeyDown][kKeyCodeLeft]			= kNPgraphRotateLeft;
	key->map[kKeyUp][kKeyCodeLeft]				= kNPgraphRotateLeftOff;

	key->map[kKeyDown][kKeyCodeRight]			= kNPgraphRotateRight;
	key->map[kKeyUp][kKeyCodeRight]				= kNPgraphRotateRightOff;

	key->map[kKeyDown][kKeyCodeUp]				= kNPgraphRotateUp;
	key->map[kKeyUp][kKeyCodeUp]				= kNPgraphRotateUpOff;

	key->map[kKeyDown][kKeyCodeDown]			= kNPgraphRotateDown;
	key->map[kKeyUp][kKeyCodeDown]				= kNPgraphRotateDownOff;

//	key->map[kKeyDown][kKeyCodeLeft]			= kNPgraphRotateCCW;
//	key->map[kKeyUp][kKeyCodeLeft]				= kNPgraphRotateCCWOff;

//	key->map[kKeyDown][kKeyCodeRight]			= kNPgraphRotateCW;
//	key->map[kKeyUp][kKeyCodeRight]				= kNPgraphRotateCWOff;

	// scale
	key->map[kKeyDown][kKeyCodeZ]				= kNPgraphZoomOn;
	key->map[kKeyUp][kKeyCodeZ]					= kNPgraphZoomOff;
}


//------------------------------------------------------------------------------
void npCloseKey (void* dataRef)
{
	return;
}

//------------------------------------------------------------------------------
void npKeyEvent (int key, int keyEventType)
{
	int command = 0;

	pData data = (pData) npGetDataRef();

	command = data->io.key.map[keyEventType][key];

	if (command != 0)
		npCtrlCommand (command, data);
}

//------------------------------------------------------------------------------
// Process key events from GLUT
void npKeyGlut (int key, int x, int y, int keyTypeGlut, int modifiers)
{
	pData data = (pData) npGetDataRef();


	if (modifiers == GLUT_ACTIVE_SHIFT)		// move these GLUT commands to npGlut, debug zz
		data->io.key.modShift = true;
	else
		data->io.key.modShift = false;

	if (modifiers == GLUT_ACTIVE_CTRL)
		data->io.key.modCtrl = true;
	else
		data->io.key.modCtrl = false;

	if (modifiers == GLUT_ACTIVE_ALT)
		data->io.key.modAlt = true;
	else
		data->io.key.modAlt = false;


	switch (keyTypeGlut) 
	{
		case kGlutKeyDown:					// ASCII key codes converted
			switch (key) 
			{
				case 8  : npKeyEvent (kKeyCodeBackSpace, kKeyDown); break;
				case 9  : npKeyEvent (kKeyCodeTab, kKeyDown); break;
				case 13 : npKeyEvent (kKeyCodeReturn, kKeyDown); break;
				case 27 : npKeyEvent (kKeyCodeESC, kKeyDown); break;
				case 32 : npKeyEvent (kKeyCodeSpace, kKeyDown); break;
				case 33 : npKeyEvent (kKeyCode1, kKeyDown); break;
				case 34 : npKeyEvent (kKeyCodeRightQuote, kKeyDown); break;  // double "
				case 35 : npKeyEvent (kKeyCode3, kKeyDown); break;
				case 36 : npKeyEvent (kKeyCode4, kKeyDown); break;
				case 37 : npKeyEvent (kKeyCode5, kKeyDown); break;
				case 38 : npKeyEvent (kKeyCode7, kKeyDown); break;
				case 39 : npKeyEvent (kKeyCodeRightQuote, kKeyDown); break;  // single '
				case 40 : npKeyEvent (kKeyCode9, kKeyDown); break;
				case 41 : npKeyEvent (kKeyCodeZero, kKeyDown); break;
				case 42 : npKeyEvent (kKeyCode8, kKeyDown); break;
				case 43 : npKeyEvent (kKeyCodeEqual, kKeyDown); break;
				case 44 : npKeyEvent (kKeyCodeComma, kKeyDown); break;
				case 45 : npKeyEvent (kKeyCodeMinus, kKeyDown); break;
				case 46 : npKeyEvent (kKeyCodePeriod, kKeyDown); break;	
				case 47 : npKeyEvent (kKeyCodeSlash, kKeyDown); break;
				case 48 : npKeyEvent (kKeyCodeZero, kKeyDown); break;
				case 49 : npKeyEvent (kKeyCode1, kKeyDown); break;
				case 50 : npKeyEvent (kKeyCode2, kKeyDown); break;
				case 51 : npKeyEvent (kKeyCode3, kKeyDown); break;
				case 52 : npKeyEvent (kKeyCode4, kKeyDown); break;
				case 53 : npKeyEvent (kKeyCode5, kKeyDown); break;
				case 54 : npKeyEvent (kKeyCode6, kKeyDown); break;
				case 55 : npKeyEvent (kKeyCode7, kKeyDown); break;
				case 56 : npKeyEvent (kKeyCode8, kKeyDown); break;
				case 57 : npKeyEvent (kKeyCode9, kKeyDown); break;
				case 58 : npKeyEvent (kKeyCodeSemiColon, kKeyDown); break;
				case 59 : npKeyEvent (kKeyCodeSemiColon, kKeyDown); break;
				case 60 : npKeyEvent (kKeyCodeComma, kKeyDown); break;
				case 61 : npKeyEvent (kKeyCodeEqual, kKeyDown); break;
				case 62 : npKeyEvent (kKeyCodePeriod, kKeyDown); break;	
				case 63 : npKeyEvent (kKeyCodeSlash, kKeyDown); break;
				case 64 : npKeyEvent (kKeyCode2, kKeyDown); break;
				case 65 : npKeyEvent (kKeyCodeA, kKeyDown); break;
				case 66 : npKeyEvent (kKeyCodeB, kKeyDown); break;
				case 67 : npKeyEvent (kKeyCodeC, kKeyDown); break;
				case 68 : npKeyEvent (kKeyCodeD, kKeyDown); break;
				case 69 : npKeyEvent (kKeyCodeE, kKeyDown); break;
				case 70 : npKeyEvent (kKeyCodeF, kKeyDown); break;
				case 71 : npKeyEvent (kKeyCodeG, kKeyDown); break;
				case 72 : npKeyEvent (kKeyCodeH, kKeyDown); break;
				case 73 : npKeyEvent (kKeyCodeI, kKeyDown); break;
				case 74 : npKeyEvent (kKeyCodeJ, kKeyDown); break;
				case 75 : npKeyEvent (kKeyCodeK, kKeyDown); break;
				case 76 : npKeyEvent (kKeyCodeL, kKeyDown); break;
				case 77 : npKeyEvent (kKeyCodeM, kKeyDown); break;
				case 78 : npKeyEvent (kKeyCodeN, kKeyDown); break;
				case 79 : npKeyEvent (kKeyCodeO, kKeyDown); break;
				case 80 : npKeyEvent (kKeyCodeP, kKeyDown); break;
				case 81 : npKeyEvent (kKeyCodeQ, kKeyDown); break;
				case 82 : npKeyEvent (kKeyCodeR, kKeyDown); break;
				case 83 : npKeyEvent (kKeyCodeS, kKeyDown); break;
				case 84 : npKeyEvent (kKeyCodeT, kKeyDown); break;
				case 85 : npKeyEvent (kKeyCodeU, kKeyDown); break;
				case 86 : npKeyEvent (kKeyCodeV, kKeyDown); break;
				case 87 : npKeyEvent (kKeyCodeW, kKeyDown); break;
				case 88 : npKeyEvent (kKeyCodeX, kKeyDown); break;
				case 89 : npKeyEvent (kKeyCodeY, kKeyDown); break;
				case 90 : npKeyEvent (kKeyCodeZ, kKeyDown); break;
				case 91 : npKeyEvent (kKeyCodeLeftBracket, kKeyDown); break;
				case 92 : npKeyEvent (kKeyCodeBackSlash, kKeyDown); break;
				case 93 : npKeyEvent (kKeyCodeRightBracket, kKeyDown); break;
				case 94 : npKeyEvent (kKeyCode6, kKeyDown); break;	
				case 95 : npKeyEvent (kKeyCodeMinus, kKeyDown); break;
				case 96 : npKeyEvent (kKeyCodeTilda, kKeyDown); break;
				case 97 : npKeyEvent (kKeyCodeA, kKeyDown); break;
				case 98 : npKeyEvent (kKeyCodeB, kKeyDown); break;
				case 99 : npKeyEvent (kKeyCodeC, kKeyDown); break;
				case 100 : npKeyEvent (kKeyCodeD, kKeyDown); break;
				case 101 : npKeyEvent (kKeyCodeE, kKeyDown); break;
				case 102 : npKeyEvent (kKeyCodeF, kKeyDown); break;
				case 103 : npKeyEvent (kKeyCodeG, kKeyDown); break;
				case 104 : npKeyEvent (kKeyCodeH, kKeyDown); break;
				case 105 : npKeyEvent (kKeyCodeI, kKeyDown); break;
				case 106 : npKeyEvent (kKeyCodeJ, kKeyDown); break;
				case 107 : npKeyEvent (kKeyCodeK, kKeyDown); break;
				case 108 : npKeyEvent (kKeyCodeL, kKeyDown); break;
				case 109 : npKeyEvent (kKeyCodeM, kKeyDown); break;
				case 110 : npKeyEvent (kKeyCodeN, kKeyDown); break;
				case 111 : npKeyEvent (kKeyCodeO, kKeyDown); break;
				case 112 : npKeyEvent (kKeyCodeP, kKeyDown); break;
				case 113 : npKeyEvent (kKeyCodeQ, kKeyDown); break;
				case 114 : npKeyEvent (kKeyCodeR, kKeyDown); break;
				case 115 : npKeyEvent (kKeyCodeS, kKeyDown); break;
				case 116 : npKeyEvent (kKeyCodeT, kKeyDown); break;
				case 117 : npKeyEvent (kKeyCodeU, kKeyDown); break;
				case 118 : npKeyEvent (kKeyCodeV, kKeyDown); break;
				case 119 : npKeyEvent (kKeyCodeW, kKeyDown); break;
				case 120 : npKeyEvent (kKeyCodeX, kKeyDown); break;
				case 121 : npKeyEvent (kKeyCodeY, kKeyDown); break;
				case 122 : npKeyEvent (kKeyCodeZ, kKeyDown); break;
				case 123 : npKeyEvent (kKeyCodeLeftBracket, kKeyDown); break;
				case 124 : npKeyEvent (kKeyCodeBackSlash, kKeyDown); break;
				case 125 : npKeyEvent (kKeyCodeRightBracket, kKeyDown); break;
				case 126 : npKeyEvent (kKeyCodeTilda, kKeyDown); break;
				case 127 : npKeyEvent (kKeyCodeDelete, kKeyDown); break;
				default : break;
			}
			break;
		case kGlutKeyUp:					// ASCII key codes converted
			switch (key) 
			{
				case 8  : npKeyEvent (kKeyCodeBackSpace, kKeyUp); break;
				case 9  : npKeyEvent (kKeyCodeTab, kKeyUp); break;
				case 13 : npKeyEvent (kKeyCodeReturn, kKeyUp); break;
				case 27 : npKeyEvent (kKeyCodeESC, kKeyUp); break;
				case 32 : npKeyEvent (kKeyCodeSpace, kKeyUp); break;
				case 33 : npKeyEvent (kKeyCode1, kKeyUp); break;
				case 34 : npKeyEvent (kKeyCodeRightQuote, kKeyUp); break;
				case 35 : npKeyEvent (kKeyCode3, kKeyUp); break;
				case 36 : npKeyEvent (kKeyCode4, kKeyUp); break;
				case 37 : npKeyEvent (kKeyCode5, kKeyUp); break;
				case 38 : npKeyEvent (kKeyCode7, kKeyUp); break;
				case 39 : npKeyEvent (kKeyCodeRightQuote, kKeyUp); break;
				case 40 : npKeyEvent (kKeyCode9, kKeyUp); break;
				case 41 : npKeyEvent (kKeyCodeZero, kKeyUp); break;
				case 42 : npKeyEvent (kKeyCode8, kKeyUp); break;
				case 43 : npKeyEvent (kKeyCodeEqual, kKeyUp); break;
				case 44 : npKeyEvent (kKeyCodeComma, kKeyUp); break;
				case 45 : npKeyEvent (kKeyCodeMinus, kKeyUp); break;
				case 46 : npKeyEvent (kKeyCodePeriod, kKeyUp); break;	
				case 47 : npKeyEvent (kKeyCodeSlash, kKeyUp); break;
				case 48 : npKeyEvent (kKeyCodeZero, kKeyUp); break;
				case 49 : npKeyEvent (kKeyCode1, kKeyUp); break;
				case 50 : npKeyEvent (kKeyCode2, kKeyUp); break;
				case 51 : npKeyEvent (kKeyCode3, kKeyUp); break;
				case 52 : npKeyEvent (kKeyCode4, kKeyUp); break;
				case 53 : npKeyEvent (kKeyCode5, kKeyUp); break;
				case 54 : npKeyEvent (kKeyCode6, kKeyUp); break;
				case 55 : npKeyEvent (kKeyCode7, kKeyUp); break;
				case 56 : npKeyEvent (kKeyCode8, kKeyUp); break;
				case 57 : npKeyEvent (kKeyCode9, kKeyUp); break;
				case 58 : npKeyEvent (kKeyCodeSemiColon, kKeyUp); break;
				case 59 : npKeyEvent (kKeyCodeSemiColon, kKeyUp); break;
				case 60 : npKeyEvent (kKeyCodeComma, kKeyUp); break;
				case 61 : npKeyEvent (kKeyCodeEqual, kKeyUp); break;
				case 62 : npKeyEvent (kKeyCodePeriod, kKeyUp); break;	
				case 63 : npKeyEvent (kKeyCodeSlash, kKeyUp); break;
				case 64 : npKeyEvent (kKeyCode2, kKeyUp); break;
				case 65 : npKeyEvent (kKeyCodeA, kKeyUp); break;
				case 66 : npKeyEvent (kKeyCodeB, kKeyUp); break;
				case 67 : npKeyEvent (kKeyCodeC, kKeyUp); break;
				case 68 : npKeyEvent (kKeyCodeD, kKeyUp); break;
				case 69 : npKeyEvent (kKeyCodeE, kKeyUp); break;
				case 70 : npKeyEvent (kKeyCodeF, kKeyUp); break;
				case 71 : npKeyEvent (kKeyCodeG, kKeyUp); break;
				case 72 : npKeyEvent (kKeyCodeH, kKeyUp); break;
				case 73 : npKeyEvent (kKeyCodeI, kKeyUp); break;
				case 74 : npKeyEvent (kKeyCodeJ, kKeyUp); break;
				case 75 : npKeyEvent (kKeyCodeK, kKeyUp); break;
				case 76 : npKeyEvent (kKeyCodeL, kKeyUp); break;
				case 77 : npKeyEvent (kKeyCodeM, kKeyUp); break;
				case 78 : npKeyEvent (kKeyCodeN, kKeyUp); break;
				case 79 : npKeyEvent (kKeyCodeO, kKeyUp); break;
				case 80 : npKeyEvent (kKeyCodeP, kKeyUp); break;
				case 81 : npKeyEvent (kKeyCodeQ, kKeyUp); break;
				case 82 : npKeyEvent (kKeyCodeR, kKeyUp); break;
				case 83 : npKeyEvent (kKeyCodeS, kKeyUp); break;
				case 84 : npKeyEvent (kKeyCodeT, kKeyUp); break;
				case 85 : npKeyEvent (kKeyCodeU, kKeyUp); break;
				case 86 : npKeyEvent (kKeyCodeV, kKeyUp); break;
				case 87 : npKeyEvent (kKeyCodeW, kKeyUp); break;
				case 88 : npKeyEvent (kKeyCodeX, kKeyUp); break;
				case 89 : npKeyEvent (kKeyCodeY, kKeyUp); break;
				case 90 : npKeyEvent (kKeyCodeZ, kKeyUp); break;
				case 91 : npKeyEvent (kKeyCodeLeftBracket, kKeyUp); break;
				case 92 : npKeyEvent (kKeyCodeBackSlash, kKeyUp); break;
				case 93 : npKeyEvent (kKeyCodeRightBracket, kKeyUp); break;
				case 94 : npKeyEvent (kKeyCode6, kKeyUp); break;	
				case 95 : npKeyEvent (kKeyCodeMinus, kKeyUp); break;
				case 96 : npKeyEvent (kKeyCodeTilda, kKeyUp); break;
				case 97 : npKeyEvent (kKeyCodeA, kKeyUp); break;
				case 98 : npKeyEvent (kKeyCodeB, kKeyUp); break;
				case 99 : npKeyEvent (kKeyCodeC, kKeyUp); break;
				case 100 : npKeyEvent (kKeyCodeD, kKeyUp); break;
				case 101 : npKeyEvent (kKeyCodeE, kKeyUp); break;
				case 102 : npKeyEvent (kKeyCodeF, kKeyUp); break;
				case 103 : npKeyEvent (kKeyCodeG, kKeyUp); break;
				case 104 : npKeyEvent (kKeyCodeH, kKeyUp); break;
				case 105 : npKeyEvent (kKeyCodeI, kKeyUp); break;
				case 106 : npKeyEvent (kKeyCodeJ, kKeyUp); break;
				case 107 : npKeyEvent (kKeyCodeK, kKeyUp); break;
				case 108 : npKeyEvent (kKeyCodeL, kKeyUp); break;
				case 109 : npKeyEvent (kKeyCodeM, kKeyUp); break;
				case 110 : npKeyEvent (kKeyCodeN, kKeyUp); break;
				case 111 : npKeyEvent (kKeyCodeO, kKeyUp); break;
				case 112 : npKeyEvent (kKeyCodeP, kKeyUp); break;
				case 113 : npKeyEvent (kKeyCodeQ, kKeyUp); break;
				case 114 : npKeyEvent (kKeyCodeR, kKeyUp); break;
				case 115 : npKeyEvent (kKeyCodeS, kKeyUp); break;
				case 116 : npKeyEvent (kKeyCodeT, kKeyUp); break;
				case 117 : npKeyEvent (kKeyCodeU, kKeyUp); break;
				case 118 : npKeyEvent (kKeyCodeV, kKeyUp); break;
				case 119 : npKeyEvent (kKeyCodeW, kKeyUp); break;
				case 120 : npKeyEvent (kKeyCodeX, kKeyUp); break;
				case 121 : npKeyEvent (kKeyCodeY, kKeyUp); break;
				case 122 : npKeyEvent (kKeyCodeZ, kKeyUp); break;
				case 123 : npKeyEvent (kKeyCodeLeftBracket, kKeyUp); break;
				case 124 : npKeyEvent (kKeyCodeBackSlash, kKeyUp); break;
				case 125 : npKeyEvent (kKeyCodeRightBracket, kKeyUp); break;
				case 126 : npKeyEvent (kKeyCodeTilda, kKeyUp); break;
				case 127 : npKeyEvent (kKeyCodeDelete, kKeyUp); break;
				default : break;
			}
			break;
		case kGlutKeyDownSpecial:			// non-ASCII Glut constants
			switch (key) 
			{
				case GLUT_KEY_F1 : npKeyEvent (kKeyCodeF1, kKeyDown); break;
				case GLUT_KEY_F2 : npKeyEvent (kKeyCodeF2, kKeyDown); break;
				case GLUT_KEY_F3 : npKeyEvent (kKeyCodeF3, kKeyDown); break;
				case GLUT_KEY_F4 : npKeyEvent (kKeyCodeF4, kKeyDown); break;
				case GLUT_KEY_F5 : npKeyEvent (kKeyCodeF5, kKeyDown); break;
				case GLUT_KEY_F6 : npKeyEvent (kKeyCodeF6, kKeyDown); break;
				case GLUT_KEY_F7 : npKeyEvent (kKeyCodeF7, kKeyDown); break;
				case GLUT_KEY_F8 : npKeyEvent (kKeyCodeF8, kKeyDown); break;
//				case GLUT_KEY_F9 : npKeyEvent (kKeyCodeF9, kKeyDown); break;
//				case GLUT_KEY_F10 : npKeyEvent (kKeyCodeF10, kKeyDown); break;
//				case GLUT_KEY_F11 : npKeyEvent (kKeyCodeF11, kKeyDown); break;
//				case GLUT_KEY_F12 : npKeyEvent (kKeyCodeF12, kKeyDown); break;
				case GLUT_KEY_LEFT : npKeyEvent (kKeyCodeLeft, kKeyDown); break;
				case GLUT_KEY_RIGHT : npKeyEvent (kKeyCodeRight, kKeyDown); break;
				case GLUT_KEY_UP : npKeyEvent (kKeyCodeUp, kKeyDown); break;
				case GLUT_KEY_DOWN : npKeyEvent (kKeyCodeDown, kKeyDown); break;
				case GLUT_KEY_PAGE_UP : npKeyEvent (kKeyCodePageUp, kKeyDown); break;
				case GLUT_KEY_PAGE_DOWN : npKeyEvent (kKeyCodePageDown, kKeyDown); break;
				case GLUT_KEY_HOME : npKeyEvent (kKeyCodeHome, kKeyDown); break;
				case GLUT_KEY_END : npKeyEvent (kKeyCodeEnd, kKeyDown); break;
//				case GLUT_KEY_INSERT : npKeyEvent (kKeyCode, kKeyDown); break;	// not compatible with Mac
				default : break;
			}
			break;
		case kGlutKeyUpSpecial:				// non-ASCII Glut constants
			switch (key) 
			{
				case GLUT_KEY_F1 : npKeyEvent (kKeyCodeF1, kKeyUp); break;
				case GLUT_KEY_F2 : npKeyEvent (kKeyCodeF2, kKeyUp); break;
				case GLUT_KEY_F3 : npKeyEvent (kKeyCodeF3, kKeyUp); break;
				case GLUT_KEY_F4 : npKeyEvent (kKeyCodeF4, kKeyUp); break;
				case GLUT_KEY_F5 : npKeyEvent (kKeyCodeF5, kKeyUp); break;
				case GLUT_KEY_F6 : npKeyEvent (kKeyCodeF6, kKeyUp); break;
				case GLUT_KEY_F7 : npKeyEvent (kKeyCodeF7, kKeyUp); break;
				case GLUT_KEY_F8 : npKeyEvent (kKeyCodeF8, kKeyUp); break;
//				case GLUT_KEY_F9 : npKeyEvent (kKeyCodeF9, kKeyUp); break;
//				case GLUT_KEY_F10 : npKeyEvent (kKeyCodeF10, kKeyUp); break;
//				case GLUT_KEY_F11 : npKeyEvent (kKeyCodeF11, kKeyUp); break;
//				case GLUT_KEY_F12 : npKeyEvent (kKeyCodeF12, kKeyUp); break;
				case GLUT_KEY_LEFT : npKeyEvent (kKeyCodeLeft, kKeyUp); break;
				case GLUT_KEY_RIGHT : npKeyEvent (kKeyCodeRight, kKeyUp); break;
				case GLUT_KEY_UP : npKeyEvent (kKeyCodeUp, kKeyUp); break;
				case GLUT_KEY_DOWN : npKeyEvent (kKeyCodeDown, kKeyUp); break;
				case GLUT_KEY_PAGE_UP : npKeyEvent (kKeyCodePageUp, kKeyUp); break;
				case GLUT_KEY_PAGE_DOWN : npKeyEvent (kKeyCodePageDown, kKeyUp); break;
				case GLUT_KEY_HOME : npKeyEvent (kKeyCodeHome, kKeyUp); break;
				case GLUT_KEY_END : npKeyEvent (kKeyCodeEnd, kKeyUp); break;
//				case GLUT_KEY_INSERT : npKeyEvent (kKeyCode, kKeyUp); break;	// not compatible with Mac
				default : break;
			}
			break;
	}
}